# TOCTOU in _get_ignore_dom (apport CVE-2019-7307)

[CVE-2019-7307](https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-2019-7307)
is a time-of-check to time-of-use (TOCTOU) vulnerability in
[apport](https://launchpad.net/ubuntu/+source/apport),
which enables an unprivileged local user to trick apport into
including the contents of an arbitrary file in a crash report.
The full bug report is public on `bugs.launchpad.net`:
[bug 1830858](https://bugs.launchpad.net/ubuntu/+source/apport/+bug/1830858).

## Instructions

I usually try to provide a `Dockerfile` so that my PoCs are safely reproducible on a patched system.
Unfortunately, apport is specifically designed to behave differently inside a container,
so I am not able to do so this time.
Instead, if you would like to test the exploit,
then you will need to revert the bug fix in your apport installation.
You can do that as follows:

```bash
git clone https://git.launchpad.net/ubuntu/+source/apport
cd apport
git checkout applied/2.20.9-0ubuntu7.6
sudo cp apport/report.py /usr/lib/python3/dist-packages/apport/report.py
```

When you are done, don't forget to fix your installation:

```bash
sudo apt-get install --reinstall python3-apport
```

Build the PoC for apport CVE-2019-7307 like this:

```bash
make
```

And run it like this:

```bash
./gencrashreport /etc/shadow
```

This will create a file named `/var/crash/_usr_share_apport_apport.0.crash`,
which is owned by `root`, but also readable by `whoopsie`.
For a full exploit chain, we therefore also need a second exploit that enables us to read files as whoopsie.
I didn't have such an exploit when CVE-2019-7307 was disclosed on 2019-07-09, but I do now:
see [README_CVE-2019-15790](README_CVE-2019-15790.md).
But to keep these instructions self-contained,
it is simplest to just change the permissions of the crash report:

```bash
sudo chmod 666 /var/crash/_usr_share_apport_apport.0.crash
```

At this point, you can unpack the crash report and see that the contents of `/etc/shadow` are embedded in the `CoreDump` file:

```bash
mkdir report
apport-unpack /var/crash/_usr_share_apport_apport.0.crash report
cd report
```

Note: `apport-unpack` is a bit flaky and usually crashes with an error message like this: `['ProcEnviron', 'UserGroups'] has no binary content`. But it works well enough to extract the core dump from the report. Now use your favorite text editor to open `CoreDump` and search for the contents of `/etc/password`. Searching for the string "root:" usually works.
